/*

*/

import java.util.concurrent.locks.ReentrantLock;

/**
   This is an example of a partially synchronized
   producer/consumer pair.

   The "bounded buffer" shared by the producer
   and consumer holds only one item.

   The one item in the shared buffer is a pair
   of numbers and the two items in the pair
   should be updated atomically.

   This version uses a lock to guarantee that the
   (single) pair of numbers in the "bounded buffer"
   get updated atomically.

   But there is no synchronization that prevents the
   producer from "overflowing", or the consumer from
   "underflowing", their shared "bounded buffer".
*/
public class ProducerConsumer_mutex_ReentrantLock
{
   private static ReentrantLock lock = new ReentrantLock();

   public static void main(String[] args) throws InterruptedException
   {
      Buffer buf = new Buffer();
      Producer producer = new Producer(buf, 1);
      Consumer consumer = new Consumer(buf, 1);
      Thread t1 = new Thread(producer);
      Thread t2 = new Thread(consumer);
      t1.start();
      t2.start();

      Thread.sleep(10_000);
    //Thread.sleep(300_000); // 5 minutes

      t1.interrupt();
      t2.interrupt();

      t1.join();
      t2.join();

      System.out.println("All threads done.");
   }


   static class Buffer
   {
      public int counter1 = -1;  // Two data items that should
      public int counter2 = -1;  // be updated atomically.
   }


   static class Producer implements Runnable
   {
      private final Buffer buf;
      private final int milliSeconds;

      public Producer(Buffer buf)
      {
         this(buf, 500);
      }

      public Producer(Buffer buf, int milliSeconds)
      {
         this.buf = buf;
         this.milliSeconds = milliSeconds;
      }

      @Override
      public void run()
      {
         int counter1 = 0;
         int counter2 = 10_000;
         try
         {
            while(! Thread.currentThread().isInterrupted())
            {
               ++counter1;
               ++counter2;
               lock.lock();  // mutual exclusion
               try
               {
                  buf.counter1 = counter1;
                  Thread.yield(); // Allow race conditions to happen more often.
                  buf.counter2 = counter2;
               }
               finally
               {
                  lock.unlock();
               }
               System.err.println("\n"+counter1+", "+counter2);
               Thread.sleep(milliSeconds);
            }
         }
         catch (InterruptedException e)
         {
            e.printStackTrace();
         }
      }
   }


   static class Consumer implements Runnable
   {
      private final Buffer buf;
      private final int milliSeconds;

      public Consumer(Buffer buf)
      {
         this(buf, 500);
      }

      public Consumer(Buffer buf, int milliSeconds)
      {
         this.buf = buf;
         this.milliSeconds = milliSeconds;
      }


      public void run()
      {
         try
         {
            while(! Thread.currentThread().isInterrupted())
            {
               int counter1, counter2;
               lock.lock();  // mutual exclusion
               try
               {
                  counter1 = buf.counter1;
                  Thread.yield(); // Allow race conditions to happen more often.
                  counter2 = buf.counter2;
                  buf.counter1 = -1; // Help detect underflows.
                  buf.counter2 = -1;
               }
               finally
               {
                  lock.unlock();
               }
               if (counter2 - counter1 != 10_000)
               {
                  System.out.println("\n\t\tBINGO: "+counter1+", "+counter2);
                  System.exit(1);
               }
               else
               {
                  System.out.println("\n\t\t"+counter1+", "+counter2);
               }
               Thread.sleep(milliSeconds);
            }
         }
         catch (InterruptedException e)
         {
            e.printStackTrace();
         }
      }
   }
}
