/*

*/

import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Condition;

/**
   This is an example of a synchronized
   producer/consumer pair.

   The "bounded buffer" shared by the producer
   and consumer holds only one item.

   The one item in the shared buffer is a pair
   of numbers and the two items in the pair
   should be updated atomically.

   This version uses a lock to guarantee that the
   (single) pair of numbers in the "bounded buffer"
   get updated atomically.

   This version uses a condition variable to syncronize
   the producer with the consumer.
*/
public class ProducerConsumer_sync_ReentrantLock_Condition
{
   private static ReentrantLock lock = new ReentrantLock();
   private static Condition fullCondition = lock.newCondition();
   private static boolean full = false;

   public static void main(String[] args) throws InterruptedException
   {
      Buffer buf = new Buffer();
      Producer producer = new Producer(buf, 1);
      Consumer consumer = new Consumer(buf, 1);
      Thread t1 = new Thread(producer);
      Thread t2 = new Thread(consumer);
      Thread t3 = new Thread(producer);
      Thread t4 = new Thread(consumer);
      t1.start();
      t2.start();
      //t3.start();  // Can cause  deadlock.
      //t4.start();  // Can cause  deadlock.

      Thread.sleep(10_000);
      //Thread.sleep(300_000); // 5 minutes

      t1.interrupt();
      t2.interrupt();
      //t3.interrupt();
      //t4.interrupt();

      t1.join(); System.out.println("t1 done");
      t2.join(); System.out.println("t2 done");
      //t3.join(); System.out.println("t3 done");
      //t4.join(); System.out.println("t4 done");

      System.out.println("All threads done.");
   }


   static class Buffer
   {
      public int counter1 = -1;  // Two data items that should
      public int counter2 = -1;  // be updated atomically.
   }


   static class Producer implements Runnable
   {
      private final Buffer buf;
      private final int milliSeconds;

      public Producer(Buffer buf)
      {
         this(buf, 500);
      }

      public Producer(Buffer buf, int milliSeconds)
      {
         this.buf = buf;
         this.milliSeconds = milliSeconds;
      }

      @Override
      public void run()
      {
         int counter1 = 0;
         int counter2 = 10_000;
         try
         {
            while(! Thread.currentThread().isInterrupted())
            {
               ++counter1;
               ++counter2;
               lock.lock();  // mutual exclusion
               try
               {
                  while (full)
                  {
                     fullCondition.await();
                  }
                  buf.counter1 = counter1;
                  Thread.yield(); // Allow race conditions to happen more often.
                  buf.counter2 = counter2;
                  full = true;
                  fullCondition.signal();    // Causes deadlock problem.
                //fullCondition.signalAll(); // Solves deadlock problem.
               }
               finally
               {
                  lock.unlock();
               }
               System.err.println("\n"+counter1+", "+counter2);
               Thread.sleep(milliSeconds);
            }
         }
         catch (InterruptedException e)
         {
            e.printStackTrace();
         }
      }
   }


   static class Consumer implements Runnable
   {
      private final Buffer buf;
      private final int milliSeconds;

      public Consumer(Buffer buf)
      {
         this(buf, 500);
      }

      public Consumer(Buffer buf, int milliSeconds)
      {
         this.buf = buf;
         this.milliSeconds = milliSeconds;
      }

      @Override
      public void run()
      {
         try
         {
            while(! Thread.currentThread().isInterrupted())
            {
               int counter1, counter2;
               lock.lock();  // mutual exclusion
               try
               {
                  while (! full) // Here, "not full" means empty.
                  {
                     fullCondition.await();
                  }
                  counter1 = buf.counter1;
                  Thread.yield(); // Allow race conditions to happen more often.
                  counter2 = buf.counter2;
                  buf.counter1 = -1; // Help detect underflows.
                  buf.counter2 = -1;
                  full = false;
                  fullCondition.signal();    // Causes deadlock problem.
                //fullCondition.signalAll(); // Solves deadlock problem.
               }
               finally
               {
                  lock.unlock();
               }
               if (counter2 - counter1 != 10_000)
               {
                  System.out.println("\n\t\tBINGO: "+counter1+", "+counter2);
                  System.exit(1);
               }
               else
               {
                  System.out.println("\n\t\t"+counter1+", "+counter2);
               }
               Thread.sleep(milliSeconds);
            }
         }
         catch (InterruptedException e)
         {
            e.printStackTrace();
         }
      }
   }
}
